# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Quantization helpers common to all uniform quantization algorithms.

This file contains quantization helpers common to all uniform quantization
algorithms. The materialize_op functions require algorithm-specific logic to
produce the quantization parameters (e.g. scale, zero point) for each tensor,
which is encapsulated in get_tensor_quant_params_fn. Each algorithm is required
to implement the get_tensor_quant_params_fn with the
qtyping.GetTensorQuantParamsFuncSignature signature.
"""

from typing import Any, Optional, Sequence
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TFLOpName = qtyping.TFLOperationName
_QuantTransformation = qtyping.QuantTransformation
_OpQuantConstraint = common_utils.OpQuantConstraint
_TensorQuantParamsCache = common_utils.TensorQuantParamsCache
_ComputePrecision = qtyping.ComputePrecision
_TensorTransformationParams = qtyping.TensorTransformationParams


def check_if_quantized(tensor: Any) -> bool:
  """Checks if the tensor is quantized."""
  return (
      tensor.quantization is not None and tensor.quantization.scale is not None
  )


def check_op_quantization_config(
    op_name: _TFLOpName,
    op_quant_config: qtyping.OpQuantizationConfig,
    config_check_policy: qtyping.ConfigCheckPolicyDict,
) -> None:
  """Checks the op quantization config.

  Args:
    op_name: The name of the op.
    op_quant_config: The quantization config for the op.
    config_check_policy: The policy to check the op quantization config.

  Raises:
    ValueError: If the op quantization config is invalid.
  """
  if op_quant_config.weight_tensor_config is None:
    raise ValueError(
        "Weight tensor quantization is required for min/max uniform"
        " quantization."
    )
  if op_quant_config.weight_tensor_config.dtype != qtyping.TensorDataType.INT:
    raise ValueError(
        "Weights need to have integer type for min/max uniform quantization. If"
        " you wish to perform float casting quantization (e.g., fp16 weight"
        " only), please set algorithm key as 'float_casting'."
    )

  if op_quant_config.min_weight_elements < 0:
    raise ValueError(
        f"min_weight_elements must be non-negative for op: {op_name} with"
        f" config: {op_quant_config}."
    )

  if op_quant_config.compute_precision in [
      _ComputePrecision.INTEGER,
      _ComputePrecision.FLOAT,
  ]:
    # Use policy-based mechanism to validate op.
    common_utils.check_if_valid_op_config(
        op_name, op_quant_config, config_check_policy
    )
  common_utils.check_subchannel_config(op_name, op_quant_config)


def materialize_input(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in the virtual input op."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_output(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in the virtual output op."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_composite(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in the virtual output op."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_add(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.add."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_sub(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.sub."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_mul(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.mul."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_softmax_and_logistic(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.softmax and tfl.logistic."""
  # Hard code scales and zp values as they are hard coded in TFL kernels.
  # Softmax:
  #   https://github.com/tensorflow/tensorflow/blob/master/tensorflow/lite/kernels/activations.cc#L548
  # Logistic:
  #   https://github.com/tensorflow/tensorflow/blob/master/tensorflow/lite/kernels/activations.cc#L421
  output_activation_constraints = {
      8: qtyping.UniformQuantParams(
          num_bits=8,
          quantized_dimension=None,
          scale=np.array(1.0 / 256),
          zero_point=np.array(-128),
          symmetric=False,
      ),
      16: qtyping.UniformQuantParams(
          num_bits=16,
          quantized_dimension=None,
          scale=np.array(1.0 / 32768),
          zero_point=np.array(0),
      ),
  }

  return common_utils.materialize_op_with_output_activation_constraint(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      output_activation_constraints,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_batch_matmul(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.batch_matmul."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_embedding_lookup(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.embedding_lookup."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      inputs_to_ignore=[0],  # Lookup index does not need to be quantized.
  )


def materialize_reshape(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.reshape."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Shape tensor does not need to be quantized.
  )


def materialize_average_pool_2d(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.average_pool_2d."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
  )


def _materialize_bias_for_fc_conv_ops(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    op_tensor_params: list[_TensorTransformationParams],
    op_input_index: int = 0,
    op_weight_index: int = 1,
    op_bias_index: int = 2,
):
  """Materializes bias tensors in conv ops by updating `op_tensor_params`.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    op_tensor_params: Partially populated quantization configuration for the
      tensors associated with the op in the order of input, weight, output.
    op_input_index: Index for the input tensor in the op.
    op_weight_index: Index for the weight tensor in the op.
    op_bias_index: Index for the bias tensor in the op.
  """
  _, weight_tensor, bias_tensor, _ = (
      tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
          op_info.op,
          graph_info.subgraph_tensors,
          op_input_index,
          op_weight_index,
          op_bias_index,
      )
  )
  if bias_tensor is not None and not check_if_quantized(bias_tensor):
    bias_quant_params = None
    # Fused bias needs to be quantized for SRQ.
    # Check if SRQ.
    if (
        op_info.op_quant_config.compute_precision == _ComputePrecision.INTEGER
        and op_info.op_quant_config.activation_tensor_config is not None
    ):
      bias_content = tfl_flatbuffer_utils.get_tensor_data(
          bias_tensor,
          graph_info.buffers,
      )
      input_consumer_params = (
          op_tensor_params[op_input_index].consumers[0].parameters
      )
      weight_consumer_params = (
          op_tensor_params[op_weight_index].consumers[0].parameters
      )
      if weight_consumer_params is None and check_if_quantized(weight_tensor):
        quant_params = weight_tensor.quantization
        if op_info.op_quant_config.weight_tensor_config is None:
          raise ValueError(
              "weight_tensor_config cannot be None when weight tensor is"
              " quantized."
          )
        weight_consumer_params = qtyping.UniformQuantParams(
            num_bits=op_info.op_quant_config.weight_tensor_config.num_bits,
            scale=quant_params.scale,
            zero_point=quant_params.zeroPoint,
            quantized_dimension=quant_params.quantizedDimension,
        )
      try:
        # Bias quantization is using fixed quantization scale:
        # input_scale * weight_scale. To avoid hidden numerics error, we check
        # the quantization error in bias quantization.
        bias_quant_params = (
            uniform_quantize_tensor.symmetric_quantize_bias_tensor(
                bias_content,
                input_consumer_params,
                weight_consumer_params,
            )
        )
      except ValueError as e:
        raise ValueError(
            f"Failed to quantize bias tensor for op {op_info.op_name} with op"
            f" id {op_info.subgraph_op_index}."
        ) from e
    # We only quantize bias under SRQ. Setting is_constant=True for SRQ only
    # to avoid quantize bias for DRQ and weight-only cases.
    is_constant = (
        # Check if SRQ.
        op_info.op_quant_config.compute_precision == _ComputePrecision.INTEGER
        and op_info.op_quant_config.activation_tensor_config is not None
    )
    op_tensor_params[op_bias_index] = (
        common_utils.get_tensor_transformation_params(
            tfl_flatbuffer_utils.get_tensor_name(bias_tensor),
            op_info,
            is_inbounding_tensor=True,
            quant_params=bias_quant_params,
            is_constant=is_constant,
        )
    )


def _are_weights_too_small(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    weight_index: int,
) -> bool:
  """Checks if weights are too small to be quantized."""
  tensor = graph_info.subgraph_tensors[op_info.op.inputs[weight_index]]
  tensor_data = tfl_flatbuffer_utils.get_tensor_data(
      tensor,
      graph_info.buffers,
  )
  return (
      tensor_data is not None
      and np.size(tensor_data) < op_info.op_quant_config.min_weight_elements
  )


def materialize_slice(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.slice."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[
          1,
          2,
      ],  # Begin and size indices do not need to be quantized.
  )


def materialize_select(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.select."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
      inputs_to_ignore=[
          0,
      ],  # Condition tensor does not need to be quantized.
  )


def materialize_select_v2(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.select_v2."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
      inputs_to_ignore=[
          0,
      ],  # Condition tensor does not need to be quantized.
  )


def materialize_dynamic_update_slice(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.dynamic_update_slice."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
      inputs_to_ignore=[
          2,
      ],  # start_indices do not need to be quantized.
  )


def materialize_sum(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.sum."""
  # For 8 bits the reference kernel calls a function without input/output
  # constraints. For all others it calls a function that enforces input/output
  # scale/zero point checks. See:
  # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/lite/kernels/reduce.cc#L909
  activation_config = op_info.op_quant_config.activation_tensor_config
  if activation_config is not None and activation_config.num_bits == 8:
    constraint = _OpQuantConstraint.NO_CONSTRAIN
  else:
    constraint = _OpQuantConstraint.SAME_AS_INPUT_SCALE
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=constraint,
      inputs_to_ignore=[1],  # Axis index does not need to be quantized.
  )


def materialize_fc_conv(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
    input_index: int = 0,
    weight_index: int = 1,
    bias_index: int = 2,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in fully_connected, conv_2d and depthwise_conv_2d.

  Args:
    get_tensor_quant_params_fn: A function to get the quantization parameters
      for a tensor.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.
    input_index: Index for the input tensor in the op.
    weight_index: Index for the weight tensor in the op.
    bias_index: Index for the bias tensor in the op.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  ignored_inputs = [bias_index]  # Bias tensor is quantized separately.
  should_ignore_weight = False
  if graph_info:
    w_tensor = graph_info.subgraph_tensors[op_info.op.inputs[weight_index]]
    should_ignore_weight = check_if_quantized(w_tensor)
  if should_ignore_weight or _are_weights_too_small(
      op_info, graph_info, weight_index
  ):
    ignored_inputs.append(weight_index)

  op_tensor_params = common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      inputs_to_ignore=ignored_inputs,
  )

  _materialize_bias_for_fc_conv_ops(
      op_info,
      graph_info,
      op_tensor_params,
      op_input_index=input_index,
      op_weight_index=weight_index,
      op_bias_index=bias_index,
  )

  return op_tensor_params


def materialize_conv2d_transpose(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.conv2d_transpose.

  Args:
    get_tensor_quant_params_fn: A function to get the quantization parameters
      for a tensor.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    Quantization configuration for the tensors associated with the op (e.g.,
    weights, bias).
  """
  ignored_shape_index = 0
  weight_index = 1
  input_index = 2
  bias_index = 3

  ignored_inputs = [
      ignored_shape_index,
      bias_index,  # Bias tensor is quantized separately.
  ]
  if _are_weights_too_small(op_info, graph_info, weight_index):
    ignored_inputs.append(weight_index)

  op_tensor_params = common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      inputs_to_ignore=ignored_inputs,
  )
  if len(op_tensor_params) < 2:
    raise ValueError(
        "Materialize standard op should return at least two tensors for"
        " conv2d_transpose."
    )
  _materialize_bias_for_fc_conv_ops(
      op_info,
      graph_info,
      op_tensor_params,
      op_input_index=input_index,
      op_weight_index=weight_index,
      op_bias_index=bias_index,
  )

  return op_tensor_params


def materialize_tanh(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.tanh."""
  # Hard code scales and zero point values as they are hard coded in:
  # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/compiler/mlir/lite/ir/tfl_ops.td#L3430
  output_activation_constraints = {}
  for num_bits in [8, 16]:
    output_activation_constraints[num_bits] = qtyping.UniformQuantParams(
        num_bits=num_bits,
        quantized_dimension=None,
        scale=np.array(1.0 / (1 << (num_bits - 1))),
        zero_point=np.array(0),
        # Activation is always asymmetric for 8 bit and symmetric for 16 bits.
        symmetric=num_bits == 16,
    )
  return common_utils.materialize_op_with_output_activation_constraint(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      output_activation_constraints,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_transpose(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.transpose."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Permutation tensor does not need to be quantized.
  )


def materialize_gelu(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.gelu."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_strided_slice(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.strided_slice."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1, 2, 3],  # Ignore the begin, end, and strides tensors.
  )


def materialize_mean(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.mean."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      inputs_to_ignore=[1],  # Axis tensor does not need to be quantized.
  )


def materialize_rsqrt(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.rsqrt."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_concatenation(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.concatenation."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
  )


def materialize_split(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.split."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[0],  # Split dimension does not need to be quantized.
  )


def materialize_pad(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.pad."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Paddings tensor does not need to be quantized.
  )


def materialize_padv2(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.padv2."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
      inputs_to_ignore=[1],  # Paddings tensor does not need to be quantized.
  )


def materialize_mirror_pad(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.mirror_pad.

  Args:
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    A list of `_TensorTransformationParams` for the tensors in the op.
  """
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Paddings tensor does not need to be quantized.
  )


def materialize_space_to_depth(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.space_to_depth.

  Args:
    get_tensor_quant_params_fn: Function to get quantization parameters for the
      tensor.
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    tensor_name_to_qsv: A map of tensor name to quantization parameters.
    tensor_quant_params_cache: Cache of already computed
      `UniformQuantParams|NonLinearQuantParams` objects keyed on a tuple of the
      buffer ID and the `TensorQuantizationConfig` used to compute it.

  Returns:
    A list of `_TensorTransformationParams` for the tensors in the op.
  """
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
  )


def materialize_squared_difference(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.squared_difference."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_max_pool_2d(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.max_pool_2d."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
  )


def materialize_resize_bilinear(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.resize_bilinear."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Resize size does not need to be quantized.
  )


def materialize_resize_nearest_neighbor(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.resize_nearest_neighbor."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Resize size does not need to be quantized.
  )


def materialize_gather_nd(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.gather_nd."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Gather indices do not need to be quantized.
  )


def materialize_maximum(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.maximum."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
  )


def materialize_pack(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.pack."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
  )


def materialize_unpack(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.unpack."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
  )


def materialize_div(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.div."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_broadcast_to(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.broadcast_to."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Shape tensor does not need to be quantized.
  )


def materialize_sqrt(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.sqrt."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_hard_swish(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.hard_swish."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_gather(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.gather."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Indices do not need to be quantized.
  )


def materialize_reduce_min(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.reduce_min."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
      constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
      inputs_to_ignore=[1],  # Axis index does not need to be quantized.
  )


def materialize_equal(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.equal."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_not_equal(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.not_equal."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def materialize_relu(
    get_tensor_quant_params_fn: qtyping.GetTensorQuantParamsFuncSignature,
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    tensor_name_to_qsv: dict[str, Any],
    tensor_quant_params_cache: _TensorQuantParamsCache,
) -> list[_TensorTransformationParams]:
  """Materialize tensors in tfl.relu."""
  return common_utils.materialize_standard_op(
      op_info,
      graph_info,
      tensor_name_to_qsv,
      get_tensor_quant_params_fn,
      tensor_quant_params_cache=tensor_quant_params_cache,
  )


def _get_tensor_shape_for_blockwise(
    tensor_shape: Sequence[int], quantized_dim: int, block_size: int
) -> list[int]:
  """Get the tensor shape for blockwise quantization.

  This function splits the quantize dimension of the tensor into blocks and the
  dim/blocks. Hence, min/max of the tensor can be calculated for each block
  using existing functions.

  Args:
    tensor_shape: The original shape of the tensor.
    quantized_dim: The dimension to be quantized blockwise.
    block_size: The size of the block.

  Returns:
    The new tensor shape for calculating scale and zp for blockwise
    quantization.
  """
  new_shape = []
  for index, val in enumerate(tensor_shape):
    if index == quantized_dim:
      new_shape.append(int(val / block_size))
      new_shape.append(block_size)
    else:
      new_shape.append(val)
  return new_shape


def _reshape_data_for_blockwise(
    tensor_data: np.ndarray,
    quantized_dim: int,
    block_size: int,
) -> tuple[np.ndarray, int]:
  """Reshapes data for blockwise quantization.

  Args:
    tensor_data: The original tensor data.
    quantized_dim: The dimension to be quantized blockwise.
    block_size: The size of the block. `block_size must be a multiple of 32. `
      `The tensor quantized dimension shape must be divisible by block_size.

  Returns:
    A tuple containing the reshaped tensor data and the new reduce dimension.
  """

  # TODO: b/417508018 - create AEQ specific error class instead of
  # using generic ValueError.
  if tensor_data.shape[quantized_dim] % block_size != 0:
    raise ValueError(
        "Tensor quantization dimension must be divisible by block size for"
        " blockwise quantization."
    )
  new_shape = _get_tensor_shape_for_blockwise(
      tensor_data.shape, quantized_dim, block_size
  )
  reshaped_data = tensor_data.reshape(new_shape)
  return reshaped_data, quantized_dim + 1


def broadcast_scale_zp_for_blockwise(
    tensor_content: np.ndarray,
    quant_params: qtyping.UniformQuantParams,
) -> qtyping.UniformQuantParams:
  """Broadcasts scale and zp for blockwise quantization.

  Args:
    tensor_content: The original tensor data.
    quant_params: The quantization parameters.
      `quant_params.quantized_dimension` must be specified.
      `quant_params.block_size` must be specified and positive.

  Returns:
    The updated quantization parameters with broadcasted scale and zp for
    correct constant quantization.
  """
  if quant_params.quantized_dimension is None:
    raise ValueError("Quantized dimension must be specified.")
  if quant_params.block_size is None or quant_params.block_size <= 0:
    raise ValueError("Block size must be specified and positive.")
  quantized_dim = quant_params.quantized_dimension
  expanded_tensor_shape = _get_tensor_shape_for_blockwise(
      tensor_content.shape, quantized_dim, quant_params.block_size
  )
  expanded_scale = np.reshape(
      np.broadcast_to(
          np.expand_dims(quant_params.scale, quantized_dim + 1),
          expanded_tensor_shape,
      ),
      tensor_content.shape,
  )
  expanded_zp = np.reshape(
      np.broadcast_to(
          np.expand_dims(quant_params.zero_point, quantized_dim + 1),
          expanded_tensor_shape,
      ),
      tensor_content.shape,
  )
  return qtyping.UniformQuantParams(
      scale=expanded_scale,
      zero_point=expanded_zp,
      num_bits=quant_params.num_bits,
      symmetric=quant_params.symmetric,
      quantized_dimension=quantized_dim,
      block_size=quant_params.block_size,
  )


def init_tensor_min_max(
    tensor_data: Optional[np.ndarray],
    op_info: qtyping.OpInfo,
) -> qtyping.QSV:
  """Initialize the min/max for a tensor.

  This function initializes the min/max values for a tensor.

  Args:
    tensor_data: The tensor data.
    op_info: Aggregated information about the op (e.g., quantization config).

  Returns:
    A dictionary containing the min/max values for the tensor, or an empty
    dictionary if the tensor data is None.
  """
  weight_tensor_config = op_info.op_quant_config.weight_tensor_config
  if tensor_data is None or weight_tensor_config is None:
    return {}
  else:
    # Get reduce dimension for min/max calculation based on quantization
    # granularity.
    granularity = weight_tensor_config.granularity
    if granularity == qtyping.QuantGranularity.TENSORWISE:
      reduce_dims = None
      keep_dims = True
    elif granularity == qtyping.QuantGranularity.CHANNELWISE:
      quantized_dim = common_utils.get_weight_quantized_dim(
          op_info, tensor_data, weight_tensor_config.granularity
      )
      reduce_dims = common_utils.get_reduce_dims(
          quantized_dim, tensor_data.shape
      )
      keep_dims = True
    elif uniform_quantize_tensor.is_blockwise(granularity):
      tensor_data, reduce_dims = (
          uniform_quantize_tensor.reshape_data_for_blockwise(
              tensor_data,
              op_info.op_name,
              granularity,
          )
      )
      keep_dims = False
    else:
      raise ValueError(f"Unsupported granularity: {granularity}")
    return {
        "min": np.min(tensor_data, axis=reduce_dims, keepdims=keep_dims),
        "max": np.max(tensor_data, axis=reduce_dims, keepdims=keep_dims),
    }



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test for various transformations used by quantization toolkit."""

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import quant_insert
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("..")


class QuantInsertTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._orig_test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/insert_dequant_test.tflite"
    )
    self._model = tfl_flatbuffer_utils.read_model(self._orig_test_model_path)

  def test_quant_insert_constant(self):
    """Test quant insert lib on a constant tensor."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quant_opcode = schema_py_generated.BuiltinOperator.QUANTIZE
    # insert quant on the constant before the add node
    quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            7,
            model.operatorCodes,
            model.buffers,
            subgraph,
            -1,
            [4],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )

    # check quant op code is added to the model
    self.assertEqual(
        model.operatorCodes[0].builtinCode,
        quant_opcode,
    )

    # check new tensor is correct created
    self.assertIn(b"_quantized", subgraph.tensors[9].name)
    self.assertEqual(
        subgraph.tensors[9].type, schema_py_generated.TensorType.INT8
    )
    self.assertEqual(
        subgraph.tensors[7].type, schema_py_generated.TensorType.UINT8
    )
    # checking if consumer has the correct input
    self.assertEqual(subgraph.operators[5].inputs[0], 6)
    self.assertEqual(subgraph.operators[5].inputs[1], 9)

    # checking the inserted node has the correct input/output
    self.assertEqual(subgraph.operators[4].outputs[0], 9)
    self.assertEqual(subgraph.operators[4].inputs[0], 7)
    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[4].opcodeIndex, 0)

  def test_quant_insert_activation(self):
    """Test quant insert lib on activation tensors."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quant_opcode = schema_py_generated.BuiltinOperator.QUANTIZE
    # insert quant on the output of a conv node
    quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            4,
            model.operatorCodes,
            model.buffers,
            subgraph,
            1,
            [3],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )

    # check quant op code is added to the model
    self.assertEqual(
        model.operatorCodes[0].builtinCode,
        quant_opcode,
    )

    # check new tensor is correctly created
    self.assertIn(b"_quantized", subgraph.tensors[9].name)
    self.assertEqual(
        subgraph.tensors[9].type, schema_py_generated.TensorType.INT8
    )
    # check original source tensor is updated
    self.assertEqual(
        subgraph.tensors[4].type, schema_py_generated.TensorType.UINT8
    )

    # checking if consumer haves the correct input
    self.assertEqual(subgraph.operators[4].inputs[0], 9)
    self.assertEqual(subgraph.operators[4].inputs[1], 5)

    # checking the inserted node has the correct input/output
    self.assertEqual(subgraph.operators[3].outputs[0], 9)
    self.assertEqual(subgraph.operators[3].inputs[0], 4)
    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[3].opcodeIndex, 0)

  def test_quant_insert_constant_multiple_consumers(self):
    """Test quant insert lib on tensors with multiple consumers."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quant_opcode = schema_py_generated.BuiltinOperator.QUANTIZE
    # insert quant on the input of a conv node
    post_trans_info = quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            2,
            model.operatorCodes,
            model.buffers,
            subgraph,
            -1,
            [1, 2],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )
    self.assertEqual(post_trans_info.op_id, 1)
    self.assertEqual(post_trans_info.num_ops_added, 1)

    # check quant op code is added to the model
    self.assertEqual(
        model.operatorCodes[0].builtinCode,
        quant_opcode,
    )

    # check new tensor is correct created
    self.assertIn(b"_quantized", subgraph.tensors[9].name)
    self.assertEqual(
        subgraph.tensors[9].type, schema_py_generated.TensorType.INT8
    )
    # check original source tensor has the correct type
    self.assertEqual(
        subgraph.tensors[2].type, schema_py_generated.TensorType.UINT8
    )

    # checking the inserted node has the correct input/output
    self.assertEqual(subgraph.operators[1].outputs[0], 9)
    self.assertEqual(subgraph.operators[1].inputs[0], 2)
    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[1].opcodeIndex, 0)

    # checking if consumer haves the correct input
    self.assertEqual(subgraph.operators[2].inputs[1], 9)
    self.assertEqual(subgraph.operators[3].inputs[1], 9)

  def test_quant_insert_activation_multiple_consumers(self):
    """Test quant insert lib on tensors with multiple consumers."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quant_opcode = schema_py_generated.BuiltinOperator.QUANTIZE
    # insert quant on the output of a conv node
    quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            1,
            model.operatorCodes,
            model.buffers,
            subgraph,
            0,
            [1, 2],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )

    # check quant op code is added to the model
    self.assertEqual(
        model.operatorCodes[0].builtinCode,
        quant_opcode,
    )

    # check new tensor is correct created
    self.assertIn(b"_quantized", subgraph.tensors[9].name)
    self.assertEqual(
        subgraph.tensors[9].type, schema_py_generated.TensorType.INT8
    )
    # check original source tensor is updated
    self.assertEqual(
        subgraph.tensors[1].type, schema_py_generated.TensorType.UINT8
    )

    # checking the inserted node has the correct input/output
    self.assertEqual(subgraph.operators[1].outputs[0], 9)
    self.assertEqual(subgraph.operators[1].inputs[0], 1)
    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[1].opcodeIndex, 0)

    # checking if consumer haves the correct input
    self.assertEqual(subgraph.operators[2].inputs[0], 9)
    self.assertEqual(subgraph.operators[3].inputs[0], 9)

  def test_quant_insert_activation_multiple_consumers_select(self):
    """Test quant insert lib on tensors with multiple consumers but only insert for one of them."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quant_opcode = schema_py_generated.BuiltinOperator.QUANTIZE
    # insert quant on the output of a conv node
    quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            1,
            model.operatorCodes,
            model.buffers,
            subgraph,
            0,
            [1],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )

    # check quant op code is added to the model
    self.assertEqual(
        model.operatorCodes[0].builtinCode,
        quant_opcode,
    )

    # check new tensor is correct created
    self.assertIn(b"_quantized", subgraph.tensors[9].name)
    self.assertEqual(
        subgraph.tensors[9].type, schema_py_generated.TensorType.INT8
    )
    # check original source tensor is updated
    self.assertEqual(
        subgraph.tensors[1].type, schema_py_generated.TensorType.UINT8
    )

    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[1].opcodeIndex, 0)

    # checking if consumer haves the correct input
    self.assertEqual(subgraph.operators[2].inputs[0], 9)
    self.assertEqual(subgraph.operators[3].inputs[0], 1)

    # checking the inserted node has the correct input/output
    self.assertEqual(subgraph.operators[1].outputs[0], 9)
    self.assertEqual(subgraph.operators[1].inputs[0], 1)

  def test_dequant_insert_on_graph_output(self):
    """Test dequant insert lib on graph output."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    # insert dequant on the graph output
    quant_insert.insert_quant(
        transformation_utils.TransformationInput(
            8,
            model.operatorCodes,
            model.buffers,
            subgraph,
            4,
            [-1],
            qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
        )
    )
    # checking inserted node is the quant node
    self.assertEqual(subgraph.operators[5].opcodeIndex, 0)
    # check if the graph output is updated
    self.assertEqual(subgraph.outputs[0], 9)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections.abc import Sequence

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import model_validator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_litert import schema_py_generated as schema_fb  # pylint: disable=g-direct-tensorflow-import


_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig


class SharedBufferTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # This two-signature model demonstrates both constant tensor and constant
    # buffer sharing. Each signature has two consecutive fully connected (FC)
    # layers that share a weight tensor: the first signature uses
    # `arith.constant` and the second uses `arith.constant1`, demonstrating
    # constant tensor sharing. Furthermore, 'arith.constant' and
    # 'arith.constant1' share a buffer, demonstrating constant buffer sharing.
    self.float_model_path = test_utils.get_path_to_datafile(
        'models/constant_tensor_and_buffer_only_sharing_weight_fcs.tflite'
    )
    self.sig1_output_tensor_name = 'PartitionedCall:0'
    self.sig2_output_tensor_name = 'PartitionedCall_1:0'

  def _get_fc_recipe_entry(
      self, regex: str, num_bits: int, granularity: str, srq: bool = False
  ):
    op_config = {
        'weight_tensor_config': {
            'num_bits': num_bits,
            'symmetric': True,
            'granularity': granularity,
            'dtype': 'INT',
            'block_size': 0,
        },
        'compute_precision': 'INTEGER',
        'explicit_dequantize': False,
        'skip_checks': False,
        'min_weight_elements': 0,
    }
    if srq:
      op_config['activation_tensor_config'] = {
          'num_bits': 8,
          'symmetric': False,
          'granularity': 'TENSORWISE',
          'dtype': 'INT',
      }
    return {
        'regex': regex,
        'operation': 'FULLY_CONNECTED',
        'algorithm_key': 'min_max_uniform_quantize',
        'op_config': op_config,
    }

  def _check_comparison_result(
      self,
      output_tensor_names: Sequence[str],
      comparison_result: model_validator.ComparisonResult,
      output_tolerance: float,
  ):
    tensors_results = comparison_result.get_all_tensor_results()
    for output_tensor_name in output_tensor_names:
      self.assertLess(
          tensors_results[output_tensor_name],
          output_tolerance,
      )

  @parameterized.named_parameters(
      dict(
          testcase_name='sig1_fcs_quant_sig2_fcs_no_quant',
          sig1_num_bits=8,
          sig2_num_bits=None,
          output_tolerance=1e-2,
      ),
      dict(
          testcase_name='sig1_fcs_no_quant_sig2_fcs_quant',
          sig1_num_bits=None,
          sig2_num_bits=8,
          output_tolerance=1e-2,
      ),
      dict(
          testcase_name='sig1_fcs_quant_sig2_fcs_quant_different_params',
          sig1_num_bits=8,
          sig2_num_bits=4,
          output_tolerance=0.5,
      ),
  )
  def test_quantization_succeeds_for_distinct_constant_tensors_with_shared_buffer_and_different_quantization_params(
      self, sig1_num_bits, sig2_num_bits, output_tolerance
  ):
    # This test checks a constant buffer shared by tensors `arith.constant`
    # and `arith.constant1` from FCs from two different signatures.
    qt = quantizer.Quantizer(self.float_model_path)

    sig1_fc1_regex = 'BatchMatMulV3;'
    sig1_fc2_regex = f'{self.sig1_output_tensor_name};'
    sig2_fc1_regex = 'BatchMatMulV31;'
    sig2_fc2_regex = f'{self.sig2_output_tensor_name};'
    recipe = []
    # Use same quantization params for FCs within each signature to ensure we
    # don't get a tensor sharing error.
    if sig1_num_bits is not None:
      recipe.append(
          self._get_fc_recipe_entry(
              sig1_fc1_regex, sig1_num_bits, 'CHANNELWISE'
          )
      )
      recipe.append(
          self._get_fc_recipe_entry(
              sig1_fc2_regex, sig1_num_bits, 'CHANNELWISE'
          )
      )
    if sig2_num_bits is not None:
      recipe.append(
          self._get_fc_recipe_entry(
              sig2_fc1_regex, sig2_num_bits, 'CHANNELWISE'
          )
      )
      recipe.append(
          self._get_fc_recipe_entry(
              sig2_fc2_regex, sig2_num_bits, 'CHANNELWISE'
          )
      )
    qt.load_quantization_recipe(recipe)
    quantized_model = qt.quantize().quantized_model
    self.assertIsNotNone(quantized_model)

    test_data = tfl_interpreter_utils.create_random_normal_input_data(
        quantized_model, num_samples=4
    )
    comparison_result = qt.validate(error_metrics='mse', test_data=test_data)
    self._check_comparison_result(
        output_tensor_names=[
            self.sig1_output_tensor_name,
            self.sig2_output_tensor_name,
        ],
        comparison_result=comparison_result,
        output_tolerance=output_tolerance,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='fc1_quant_fc2_no_quant',
          fc1_num_bits=8,
          fc2_num_bits=None,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.0015,
      ),
      dict(
          testcase_name='fc1_no_quant_fc2_quant',
          fc1_num_bits=None,
          fc2_num_bits=8,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.0015,
      ),
      dict(
          testcase_name='fc1_quant_fc2_quant_different_params',
          fc1_num_bits=8,
          fc2_num_bits=4,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.25,
      ),
      dict(
          testcase_name='fc1_tensorwise_fc2_channelwise',
          fc1_num_bits=8,
          fc2_num_bits=8,
          fc1_granularity='TENSORWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.25,
      ),
      dict(
          testcase_name='fc1_channelwise_fc2_tensorwise',
          fc1_num_bits=8,
          fc2_num_bits=8,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='TENSORWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.25,
      ),
      dict(
          testcase_name='fc1_tensorwise_fc2_tensorwise_different_bits',
          fc1_num_bits=8,
          fc2_num_bits=4,
          fc1_granularity='TENSORWISE',
          fc2_granularity='TENSORWISE',
          fc1_srq=False,
          fc2_srq=False,
          output_tolerance=0.30,
      ),
      dict(
          testcase_name='fc1_srq_fc2_drq',
          fc1_num_bits=8,
          fc2_num_bits=8,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=True,
          fc2_srq=False,
          output_tolerance=0.60,
      ),
      dict(
          testcase_name='fc1_drq_fc2_srq',
          fc1_num_bits=8,
          fc2_num_bits=8,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='CHANNELWISE',
          fc1_srq=False,
          fc2_srq=True,
          output_tolerance=0.60,
      ),
      dict(
          testcase_name='fc1_srq_fc2_srq',
          fc1_num_bits=8,
          fc2_num_bits=8,
          fc1_granularity='CHANNELWISE',
          fc2_granularity='TENSORWISE',
          fc1_srq=True,
          fc2_srq=True,
          output_tolerance=0.60,
      ),
  )
  def test_quantization_succeeds_for_a_constant_tensor_with_different_quantization_params(
      self,
      fc1_num_bits,
      fc2_num_bits,
      output_tolerance,
      fc1_granularity,
      fc2_granularity,
      fc1_srq,
      fc2_srq,
  ):
    # This test checks a constant tensor `arith.constant` shared by FCs in the
    # first signature.
    qt = quantizer.Quantizer(self.float_model_path)

    sig1_fc1_regex = 'BatchMatMulV3'
    sig1_fc2_regex = 'PartitionedCall:0'

    recipe = []
    if fc1_num_bits is not None:
      recipe.append(
          self._get_fc_recipe_entry(
              sig1_fc1_regex, fc1_num_bits, fc1_granularity, srq=fc1_srq
          )
      )
    if fc2_num_bits is not None:
      recipe.append(
          self._get_fc_recipe_entry(
              sig1_fc2_regex, fc2_num_bits, fc2_granularity, srq=fc2_srq
          )
      )
    qt.load_quantization_recipe(recipe)

    calibration_result = None
    if qt.need_calibration:
      calib_data = tfl_interpreter_utils.create_random_normal_input_data(
          self.float_model_path, num_samples=5
      )
      calibration_result = qt.calibrate(calib_data)
    quantized_model = qt.quantize(
        calibration_result=calibration_result
    ).quantized_model
    self.assertIsNotNone(quantized_model)

    model = schema_fb.Model.GetRootAsModel(quantized_model, 0)
    subgraph = model.Subgraphs(0)
    fc_ops = []
    for i in range(subgraph.OperatorsLength()):
      op = subgraph.Operators(i)
      op_code = model.OperatorCodes(op.OpcodeIndex())
      if op_code.BuiltinCode() == schema_fb.BuiltinOperator.FULLY_CONNECTED:
        fc_ops.append(op)

    sig1_fc1_op, sig1_fc2_op = None, None
    for op in fc_ops:
      output_tensor = subgraph.Tensors(op.Outputs(0))
      if output_tensor.Name() == b'PartitionedCall:0':
        sig1_fc2_op = op
      elif output_tensor.Name() == b'BatchMatMulV3':
        sig1_fc1_op = op

    self.assertIsNotNone(sig1_fc1_op)
    self.assertIsNotNone(sig1_fc2_op)

    if fc1_num_bits is not None:
      fc1_weight = subgraph.Tensors(sig1_fc1_op.Inputs(1))
      self.assertIsNotNone(fc1_weight.Quantization())
      self.assertIsNotNone(fc1_weight.Quantization().ScaleAsNumpy())
      if fc1_granularity == 'TENSORWISE':
        self.assertEqual(fc1_weight.Quantization().ScaleAsNumpy().size, 1)
      elif fc1_granularity == 'CHANNELWISE':
        self.assertEqual(fc1_weight.Quantization().ScaleAsNumpy().size, 4)

    if fc2_num_bits is not None:
      fc2_weight = subgraph.Tensors(sig1_fc2_op.Inputs(1))
      self.assertIsNotNone(fc2_weight.Quantization())
      self.assertIsNotNone(fc2_weight.Quantization().ScaleAsNumpy())
      if fc2_granularity == 'TENSORWISE':
        self.assertEqual(fc2_weight.Quantization().ScaleAsNumpy().size, 1)
      elif fc2_granularity == 'CHANNELWISE':
        self.assertEqual(fc2_weight.Quantization().ScaleAsNumpy().size, 4)

    test_data = tfl_interpreter_utils.create_random_normal_input_data(
        quantized_model, num_samples=4
    )
    comparison_result = qt.validate(error_metrics='mse', test_data=test_data)
    self._check_comparison_result(
        output_tensor_names=[
            self.sig1_output_tensor_name,
            self.sig2_output_tensor_name,
        ],
        comparison_result=comparison_result,
        output_tolerance=output_tolerance,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Apply dequantization transformations to the given op/tensor.

Inserts dequantize node after the given tensor to enable float execution of
the tensor consumer
"""

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import quantize_tensor
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


def insert_dequant(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Insert dequant op after the given tensor in the subgraph.

  Args:
    transformation_input: input structure that contains all information needed
      for the transformation.

  Returns:
    TransformationInfo:
      op_id: the index where the dequant op is added
      num_ops_added: the total number of ops inserted by this operation, which
        is 1
  """
  dequant_op_code_idx = transformation_utils.add_op_code(
      schema_py_generated.BuiltinOperator.DEQUANTIZE,
      transformation_input.op_codes,
  )
  # create output tensor for the dequant op
  tensor = transformation_input.subgraph.tensors[transformation_input.tensor_id]
  new_tensor_id = transformation_utils.add_new_activation_tensor(
      tensor.name + b'_dequant',
      tensor.shape,
      schema_py_generated.TensorType.FLOAT32,
      transformation_input.subgraph,
  )

  # create dequantize_op
  dequant_op = schema_py_generated.OperatorT()
  dequant_op.opcodeIndex = dequant_op_code_idx
  dequant_op.outputs = [new_tensor_id]
  dequant_op.inputs = [transformation_input.tensor_id]

  # quantize the source tensor
  quantize_tensor.quantize_tensor(transformation_input)

  # update the original consumers of the op to take the dequant op,
  # and find the first consumer of the new tensor
  first_consumer_id = min(transformation_input.consumers)
  for consumer_id in transformation_input.consumers:
    op = transformation_input.subgraph.operators[consumer_id]
    for input_idx in range(len(op.inputs)):
      if op.inputs[input_idx] == transformation_input.tensor_id:
        op.inputs[input_idx] = new_tensor_id

  # if the output is also an output to the graph, we need to update that as well
  for output_idx, output in enumerate(transformation_input.subgraph.outputs):
    if output == transformation_input.tensor_id:
      transformation_input.subgraph.outputs[output_idx] = new_tensor_id

  # add dequant into the subgraph op list,
  # must insert the op right before it's first consumer
  # in the case of output goes to graph output, we need to ensure the dequant
  # op is inserted after the producer
  op_id = max(transformation_input.producer + 1, first_consumer_id)
  transformation_input.subgraph.operators.insert(op_id, dequant_op)
  return qtyping.TransformationInfo(
      op_id=op_id, num_ops_added=1, output_tensor_id=new_tensor_id
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for tensor_utils."""

import dataclasses

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor

_IntType = uniform_quantize_tensor.IntType


class TensorUtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._test_data = np.array([[1, 2], [4, 5]])
    self._expected_quantized_result = np.array(
        [[9, 19], [22, 27]], dtype=np.int8
    )

  @parameterized.parameters(
      (_IntType(8, True), (-128, 127)),
      (_IntType(8, False), (0, 255)),
      (_IntType(4, True), (-8, 7)),
      (_IntType(4, False), (0, 15)),
      (_IntType(2, True), (-2, 1)),
      (_IntType(2, False), (0, 3)),
  )
  def test_get_quantized_range(self, qtype, expected_range):
    self.assertEqual(
        expected_range, uniform_quantize_tensor.get_quantized_range(qtype)
    )

  @parameterized.parameters(
      (_IntType(8, True), np.int8),
      (_IntType(16, False), np.uint16),
      (_IntType(32, True), np.int32),
      (_IntType(64, True), np.int64),
  )
  def test_assign_quantized_type(self, qtype, expected_type):
    sample_input = np.array([2.0, 2.0])
    result = uniform_quantize_tensor.assign_quantized_type(sample_input, qtype)
    self.assertEqual(expected_type, result.dtype)

  @parameterized.named_parameters(
      dict(
          testcase_name="scalar_tensor_1d_params",
          scale=np.array([0.1]),
          zero_point=np.array([-1], dtype=np.int8),
          tensor_data=np.array(6.66),
          quantized_dimension=None,
      ),
      dict(
          testcase_name="1d_tensor_scalar_params",
          scale=np.array(0.1),
          zero_point=np.array(-1, dtype=np.int8),
          tensor_data=np.array([6.66, 8.88]),
          quantized_dimension=None,
      ),
      dict(
          testcase_name="2d_tensor_1d_params",
          scale=np.array([0.1, 0.2]),
          zero_point=np.array([-1, 2], dtype=np.int8),
          tensor_data=np.array([[1, 2], [4, 5]]),
          quantized_dimension=0,
      ),
  )
  def test_fix_quantization_params_rank_succeed(
      self, scale, zero_point, tensor_data, quantized_dimension
  ):
    quant_params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=quantized_dimension,
        scale=scale,
        zero_point=zero_point,
        symmetric=False,
    )
    fixed_quant_params = uniform_quantize_tensor.fix_quantization_params_rank(
        tensor_data, quant_params
    )
    self.assertEqual(fixed_quant_params.scale.ndim, tensor_data.ndim)
    self.assertEqual(fixed_quant_params.zero_point.ndim, tensor_data.ndim)

  def test_fix_quantization_params_rank_scalar_tensor_1d_params_raise(self):
    quant_params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.array([0.1, 0.5]),
        zero_point=np.array([-1, -2], dtype=np.int8),
        symmetric=False,
    )
    error_message = (
        "Scale and zero_point must contain single element for scalar tensor."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.fix_quantization_params_rank(
          np.array(6.66), quant_params
      )

  @parameterized.parameters(
      (
          [-3.0, 1.3, 2.4, 16.0],
          [0.12598425],
          [0],
          8,
          False,
          [-24, 10, 19, 127],
      ),
      (
          [-16.0, 1.3, 2.4, 16.0],
          [0.12598425],
          [0],
          8,
          True,
          [-127, 10, 19, 127],  # int8 symmetric is narrow range, -127 to 127
      ),
      (
          [-3.0, 1.3, 2.4, 16.0],
          [1.2666667],
          [-6],
          4,
          False,
          [-8, -5, -4, 7],
      ),
      (
          [-3.0, 1.3, 2.4, 16.0],
          [1.2666667],
          [-6],
          4,
          True,
          [-8, -5, -4, 7],  # int4 symmetric is not narrow range, -8 to 7
      ),
  )
  def test_uniform_quantize(
      self, tensor, scale, zero_points, num_bits, symmetric, expected_tensor
  ):
    quant_params = qtyping.UniformQuantParams(
        quantized_dimension=0,
        num_bits=num_bits,
        scale=np.array(scale),
        zero_point=np.array(zero_points),
        symmetric=symmetric,
    )

    quantized_tensor = uniform_quantize_tensor.uniform_quantize(
        np.array(tensor), quant_params
    )

    self.assertSequenceAlmostEqual(expected_tensor, quantized_tensor)

  def test_uniform_quantize_wrong_shape(self):
    tensor = [-3.0, 1.3, 2.4, 16.0]

    error_message = (
        "Ranks of scales (3) and zps (2) must be the same as the tensor rank"
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.uniform_quantize(
          np.array(tensor),
          qtyping.UniformQuantParams(
              quantized_dimension=0,
              num_bits=4,
              scale=np.array([[[1.2666667]]]),
              zero_point=np.array([[-6]]),
              symmetric=True,
          ),
      )

    error_message = "Ranks of scales"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.uniform_quantize(
          np.array(tensor),
          qtyping.UniformQuantParams(
              quantized_dimension=0,
              num_bits=4,
              scale=np.array([[1.2666667]]),
              zero_point=np.array([[-6]]),
              symmetric=True,
          ),
      )

  def test_uniform_quantize_quant_dim_not_divisible_by_block_size_raise(self):
    tensor = np.random.rand(34, 2)
    error_message = (
        "Quantized dimension 34 in tensor shape (34, 2) is not divisible by"
        " block size 32."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.uniform_quantize(
          np.array(tensor),
          qtyping.UniformQuantParams(
              quantized_dimension=0,
              block_size=32,
              num_bits=4,
              scale=np.array([1.2666667]),
              zero_point=np.array([-6]),
              symmetric=True,
          ),
          is_blockwise_quant=True,
      )

  @parameterized.parameters(
      (
          8,
          [-24, 10, 19, 127],
          [0.12598425],
          [0],
          [-3.023622, 1.2598425, 2.3937008, 16.0],
      ),
      (
          4,
          [-8, -5, -4, 7],
          [1.2666667],
          [-6],
          [-2.5333335, 1.2666668, 2.5333335, 16.466667],
      ),
  )
  def test_uniform_dequantize(
      self,
      num_bits,
      quantized_tensor,
      scale,
      zero_points,
      expected_output_tensor,
  ):
    quant_params = qtyping.UniformQuantParams(
        quantized_dimension=0,
        num_bits=num_bits,
        scale=np.array(scale),
        zero_point=np.array(zero_points),
        symmetric=False,
    )

    dequantized_tensor = uniform_quantize_tensor.uniform_dequantize(
        np.array(quantized_tensor), quant_params
    )

    self.assertSequenceAlmostEqual(
        expected_output_tensor, dequantized_tensor, places=4
    )

  def test_uniform_dequantize_wrong_shape(self):
    tensor = [-3.0, 1.3, 2.4, 16.0]

    error_message = (
        "Ranks of scales (3) and zps (2) must be the same as the tensor rank"
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.uniform_dequantize(
          np.array(tensor),
          qtyping.UniformQuantParams(
              quantized_dimension=0,
              num_bits=4,
              scale=np.array([[[1.2666667]]]),
              zero_point=np.array([[-6]]),
              symmetric=True,
          ),
      )

    error_message = "Ranks of scales"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      uniform_quantize_tensor.uniform_dequantize(
          np.array(tensor),
          qtyping.UniformQuantParams(
              quantized_dimension=0,
              num_bits=4,
              scale=np.array([[1.2666667]]),
              zero_point=np.array([[-6]]),
              symmetric=True,
          ),
      )

  def test_uniform_dequantize_blockwise(self):
    quantized_tensor = np.array([[-8, -5, -4, 7], [-4, 7, -8, -5]])
    expected_output_tensor = np.array([
        [-10.1333336, -6.3333335, -5.0666668, 8.8666669],
        [-5.0666668, 8.8666669, -10.1333336, -6.3333335],
    ])
    quant_params = qtyping.UniformQuantParams(
        # b/443830202:
        quantized_dimension=0,
        num_bits=4,
        scale=np.array([[[1.2666667, 1.2666667], [1.2666667, 1.2666667]]]),
        zero_point=np.array([[0]]),
        symmetric=True,
        block_size=2,
    )

    dequantized_tensor = uniform_quantize_tensor.uniform_dequantize(
        np.array(quantized_tensor), quant_params
    )

    self.assertSequenceAlmostEqual(
        expected_output_tensor.flatten(), dequantized_tensor.flatten(), places=4
    )

  @parameterized.parameters(
      (8, 8, True, True),
      (8, 4, False, True),
      (16, 8, True, False),
      (16, 8, True, True),
  )
  def test_quantize_bias_tensor(
      self,
      activation_num_bits,
      weight_num_bits,
      symmetric_weights,
      channelwise_weight,
  ):
    input_quant_config = qtyping.UniformQuantParams(
        scale=np.array([0.8]),
        zero_point=np.array([10]),
        num_bits=activation_num_bits,
        symmetric=False,
        quantized_dimension=None,
    )
    weight_scale, weight_zp = [0.1], [-1]
    num_channels, quantized_dimension = 1, None
    if channelwise_weight:
      num_channels = 2
      weight_scale = weight_scale * num_channels
      weight_zp = weight_zp * num_channels
      quantized_dimension = 0

    weight_quant_config = qtyping.UniformQuantParams(
        quantized_dimension=0,
        num_bits=weight_num_bits,
        scale=np.array(weight_scale, dtype=np.float32),
        zero_point=np.array(weight_zp, dtype=np.int8),
        symmetric=symmetric_weights,
    )
    bias_tensor_data = np.array([66.0, 88.0])

    bias_quant_config = uniform_quantize_tensor.symmetric_quantize_bias_tensor(
        bias_tensor_data,
        input_quant_config,
        weight_quant_config,
    )
    bias_num_bits = 32 if activation_num_bits == 8 else 64
    self.assertEqual(bias_quant_config.num_bits, bias_num_bits)
    # Alwasys a 1D array
    self.assertLen(bias_quant_config.scale.shape, 1)
    self.assertLen(bias_quant_config.zero_point.shape, 1)

    self.assertLen(bias_quant_config.scale, num_channels)
    effective_scale = input_quant_config.scale[0] * weight_quant_config.scale[0]
    self.assertEqual(bias_quant_config.scale[0], effective_scale)
    self.assertEqual(bias_quant_config.zero_point[0], 0)  # Always symmetric
    self.assertEqual(bias_quant_config.symmetric, True)
    self.assertEqual(bias_quant_config.quantized_dimension, quantized_dimension)

    # Check quantized content
    dequantized_bias = uniform_quantize_tensor.uniform_dequantize(
        bias_quant_config.quantized_data, bias_quant_config
    )
    self.assertSequenceAlmostEqual(
        list(dequantized_bias.flatten()), list(bias_tensor_data), places=5
    )

    if activation_num_bits == 16:
      # Check if it is safe to cast int64 bias to int32. We save the int32
      # quantized bias as int64 if the input tensor is quantized to 16 bits.
      # This is to assume the matmul is using int64 accumulator (safe from
      # overflow). For accelerators with int32 accumulator, it is safe to cast
      # int64 back to int32.
      quantized_bias = bias_quant_config.quantized_data
      self.assertIsNotNone(quantized_bias)
      self.assertEqual(quantized_bias.dtype, np.int64)
      self.assertSequenceEqual(
          list(quantized_bias.flatten()),
          list(quantized_bias.astype(np.int32).flatten()),
      )

      bias_quant_config = dataclasses.replace(
          bias_quant_config,
          num_bits=32,
      )

    expected_quantized_data = uniform_quantize_tensor.uniform_quantize(
        bias_tensor_data, bias_quant_config
    )
    self.assertSequenceEqual(
        list(expected_quantized_data.flatten()),
        list(bias_quant_config.quantized_data.flatten()),  # pytype: disable=attribute-error
    )

  @parameterized.parameters(True, False)
  def test_quantize_bias_tensor_raises_error_for_large_quantization_error(
      self, check_error
  ):
    input_quant_config = qtyping.UniformQuantParams(
        scale=np.array([0.1]),
        zero_point=np.array([10]),
        num_bits=8,
        symmetric=False,
        quantized_dimension=None,
    )
    weight_quant_config = qtyping.UniformQuantParams(
        scale=np.array([0.1]),
        zero_point=np.array([-1]),
        num_bits=8,
        symmetric=True,
        quantized_dimension=None,
    )
    # This will result in quantized bias of 3e9, which is larger than int32 max.
    bias_tensor_data = np.array([3e7])

    if check_error:
      with self.assertRaisesRegex(
          ValueError,
          "Quantization error is too large for bias tensor quantization.",
      ):
        uniform_quantize_tensor.symmetric_quantize_bias_tensor(
            bias_tensor_data,
            input_quant_config,
            weight_quant_config,
            check_error
        )
    else:
      self.assertIsNotNone(
          uniform_quantize_tensor.symmetric_quantize_bias_tensor(
              bias_tensor_data,
              input_quant_config,
              weight_quant_config,
              check_error,
          )
      )

  @parameterized.parameters((8, True), (16, False))
  def test_tensor_zp_scale_from_min_max(self, num_bits, symmetric):
    min_val = np.min(self._test_data, keepdims=True)
    max_val = np.max(self._test_data, keepdims=True)

    zp, scale = uniform_quantize_tensor.tensor_zp_scale_from_min_max(
        min_val,
        max_val,
        num_bits,
        symmetric,
        qtyping.QuantGranularity.TENSORWISE,
    )
    self.assertEqual(zp.shape, scale.shape)
    max_q = 2**num_bits / 2 - 1
    calculated_max = scale[0] * (max_q - zp[0])
    self.assertAlmostEqual(calculated_max, max_val, delta=1e-3)
    min_q = -(2**num_bits) / 2
    if symmetric:
      min_q += 1
    calculated_min = scale[0] * (min_q - zp[0])
    if symmetric:
      self.assertAlmostEqual(calculated_min, -max_val, delta=1e-3)
    else:
      # Range has to be extended to include zero.
      self.assertEqual(calculated_min, 0)

  @parameterized.parameters(
      # number of bits, is_symmetric, max bound of the quantized range.
      (4, True, 7),
      (8, False, 255),
    )
  def test_tensor_zp_scale_from_min_max_with_clipping(
      self, num_bits, symmetric, quantized_bound
  ):
    min_val = np.array([[1.0]])
    max_val = np.array([[5.0]])
    clipping_values = np.array([4.0])
    zp, scale = uniform_quantize_tensor.tensor_zp_scale_from_min_max(
        min_val,
        max_val,
        num_bits,
        symmetric,
        qtyping.QuantGranularity.TENSORWISE,
        clipping_values,
    )
    expected_scale = clipping_values / quantized_bound

    with self.subTest(name="CheckShapes"):
      self.assertEqual(zp.shape, scale.shape)
      self.assertEqual(zp.shape, (1, 1))

    if symmetric:
      with self.subTest(name="CheckSymmetricZpValue"):
        self.assertEqual(zp[0], 0)

    with self.subTest(name="CheckScaleValue"):
      self.assertEqual(scale[0], expected_scale)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for algorithm_manager_api."""


import absl.testing.absltest as absltest
from absl.testing import parameterized

from ai_edge_quantizer import algorithm_manager_api
from ai_edge_quantizer import default_policy
from ai_edge_quantizer import qtyping

_TFLOpName = qtyping.TFLOperationName


# Sample functions for test cases.
def _sample_init_qsvs(*_, **__):
  return 1.0, dict()


def _sample_calibration_func(*_, **__):
  return 2.0, dict()


def _sample_materialize_func(*_, **__):
  return 3.0, dict()


def _sample_check_op_config_func(_, op_config):
  if op_config.weight_tensor_config.num_bits == 17:
    raise ValueError("Unsupported number of bits.")


class AlgorithmManagerApiTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._alg_manager = algorithm_manager_api.AlgorithmManagerApi()

  def test_register_op_quant_config_validation_func_succeeds(self):
    self.assertEmpty(self._alg_manager._config_check_registry)
    test_algorithm_name = "test_algorithm"
    self._alg_manager.register_op_quant_config_validation_func(
        test_algorithm_name, _sample_check_op_config_func
    )
    self.assertIn(test_algorithm_name, self._alg_manager._config_check_registry)
    check_func = self._alg_manager._config_check_registry[test_algorithm_name]
    self.assertEqual(check_func, _sample_check_op_config_func)

  def test_register_quantized_op(self):
    self._alg_manager.register_quantized_op(
        algorithm_key="ptq",
        tfl_op_name=_TFLOpName.FULLY_CONNECTED,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    self._alg_manager.register_quantized_op(
        algorithm_key="gptq",
        tfl_op_name=_TFLOpName.CONV_2D,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    self.assertTrue(self._alg_manager.is_algorithm_registered("ptq"))
    self.assertTrue(self._alg_manager.is_algorithm_registered("gptq"))
    self.assertTrue(
        self._alg_manager.is_op_registered("ptq", _TFLOpName.FULLY_CONNECTED)
    )
    self.assertTrue(
        self._alg_manager.is_op_registered("gptq", _TFLOpName.CONV_2D)
    )
    self.assertFalse(
        self._alg_manager.is_op_registered("gptq", _TFLOpName.DEPTHWISE_CONV_2D)
    )

  def test_get_supported_ops(self):
    algorithm_key = "ptq"
    self._alg_manager.register_quantized_op(
        algorithm_key=algorithm_key,
        tfl_op_name=_TFLOpName.FULLY_CONNECTED,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    self._alg_manager.register_quantized_op(
        algorithm_key=algorithm_key,
        tfl_op_name=_TFLOpName.CONV_2D,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    supported_ops = self._alg_manager.get_supported_ops(algorithm_key)
    self.assertIn(_TFLOpName.CONV_2D, supported_ops)
    self.assertIn(_TFLOpName.FULLY_CONNECTED, supported_ops)
    self.assertNotIn(_TFLOpName.DEPTHWISE_CONV_2D, supported_ops)

  def test_get_quantization_func(self):
    algorithm_key = "ptq"
    tfl_op = _TFLOpName.FULLY_CONNECTED
    self._alg_manager.register_quantized_op(
        algorithm_key=algorithm_key,
        tfl_op_name=tfl_op,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    materialize_func = self._alg_manager.get_quantization_func(
        algorithm_key,
        tfl_op,
        qtyping.QuantizeMode.MATERIALIZE,
    )
    self.assertEqual(_sample_materialize_func()[0], materialize_func()[0])
    calibration_func = self._alg_manager.get_quantization_func(
        algorithm_key,
        tfl_op,
        qtyping.QuantizeMode.CALIBRATE,
    )
    self.assertEqual(_sample_calibration_func()[0], calibration_func()[0])

    # Query for unsupported operation.
    error_message = "Unsupported operation"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._alg_manager.get_quantization_func(
          algorithm_key,
          _TFLOpName.BATCH_MATMUL,
          qtyping.QuantizeMode.MATERIALIZE,
      )

    # Query for unregisted algorithm.
    error_message = "Unregistered algorithm"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._alg_manager.get_quantization_func(
          "gptq",
          tfl_op,
          qtyping.QuantizeMode.MATERIALIZE,
      )

  def test_get_init_qsv_func(self):
    algorithm_key = "ptq"
    tfl_op = _TFLOpName.FULLY_CONNECTED
    self._alg_manager.register_quantized_op(
        algorithm_key=algorithm_key,
        tfl_op_name=tfl_op,
        init_qsv_func=_sample_init_qsvs,
        calibration_func=_sample_calibration_func,
        materialize_func=_sample_materialize_func,
    )
    init_qsv_func = self._alg_manager.get_init_qsv_func(algorithm_key, tfl_op)
    self.assertEqual(_sample_init_qsvs()[0], init_qsv_func()[0])

    # Query for unsupported operation.
    error_message = "Unsupported operation"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._alg_manager.get_init_qsv_func(
          algorithm_key,
          _TFLOpName.BATCH_MATMUL,
      )

    # Query for unregisted algorithm.
    error_message = "Unregistered algorithm"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._alg_manager.get_init_qsv_func(
          "gptq",
          tfl_op,
      )

  def test_register_config_check_policy_succeeds(self):
    self.assertEmpty(self._alg_manager._config_check_policy_registry)
    test_algorithm_name = "test_algorithm"
    test_config_check_policy = qtyping.ConfigCheckPolicyDict({
        _TFLOpName.FULLY_CONNECTED: {
            qtyping.OpQuantizationConfig(
                weight_tensor_config=qtyping.TensorQuantizationConfig(
                    num_bits=1
                )
            )
        }
    })
    self._alg_manager.register_config_check_policy(
        test_algorithm_name, test_config_check_policy
    )
    self.assertIn(
        test_algorithm_name, self._alg_manager._config_check_policy_registry
    )
    self.assertIsNotNone(
        self._alg_manager._config_check_policy_registry[test_algorithm_name]
    )

  def test_default_policy_not_empty(self):
    """Tests that the default policy is not empty & no empty policy is generated."""
    self.assertNotEmpty(default_policy.DEFAULT_CONFIG_CHECK_POLICY)
    for policy in default_policy.DEFAULT_CONFIG_CHECK_POLICY.values():
      self.assertNotEmpty(policy)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../tests/models")


class TflUtilsSingleSignatureModelTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    self._input_data = np.random.rand(1, 28, 28, 1).astype(np.float32)

  def test_create_tfl_interpreter(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    self.assertIsNotNone(tfl_interpreter)

  def test_invoke_interpreter_once(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    tfl_interpreter_utils.invoke_interpreter_once(
        tfl_interpreter, [self._input_data]
    )
    output_details = tfl_interpreter.get_output_details()[0]
    output_data = tfl_interpreter.get_tensor(output_details["index"])
    self.assertIsNotNone(output_data)
    self.assertEqual(tuple(output_data.shape), (1, 10))
    self.assertAlmostEqual(output_data[0][0], 0.0031010755)

  def test_get_tensor_data(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    tfl_interpreter_utils.invoke_interpreter_once(
        tfl_interpreter, [self._input_data]
    )
    output_details = tfl_interpreter.get_output_details()[0]
    output_data = tfl_interpreter_utils.get_tensor_data(
        tfl_interpreter, output_details
    )
    self.assertEqual(tuple(output_data.shape), (1, 10))

  def test_get_tensor_name_to_content_map(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    tfl_interpreter_utils.invoke_interpreter_once(
        tfl_interpreter, [self._input_data]
    )

    tensor_name_to_content_map = (
        tfl_interpreter_utils.get_tensor_name_to_content_map(tfl_interpreter)
    )
    input_content = tensor_name_to_content_map["serving_default_conv2d_input:0"]
    self.assertSequenceAlmostEqual(
        self._input_data.flatten(), input_content.flatten()
    )
    weight_content = tensor_name_to_content_map["sequential/conv2d/Conv2D"]
    self.assertEqual(tuple(weight_content.shape), (8, 3, 3, 1))

    self.assertIn(
        "sequential/average_pooling2d/AvgPool", tensor_name_to_content_map
    )
    average_pool_res = tensor_name_to_content_map[
        "sequential/average_pooling2d/AvgPool"
    ]
    self.assertEqual(tuple(average_pool_res.shape), (1, 14, 14, 8))

  def test_get_tensor_name_to_content_map_fails_no_preserve_all_tensors(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path, preserve_all_tensors=False
    )
    tfl_interpreter_utils.invoke_interpreter_once(
        tfl_interpreter, [self._input_data]
    )
    with self.assertRaisesRegex(ValueError, "Tensor data is null."):
      tfl_interpreter_utils.get_tensor_name_to_content_map(tfl_interpreter)

  def test_has_quantization_params(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    input_details = tfl_interpreter.get_input_details()[0]
    self.assertFalse(
        tfl_interpreter_utils.has_quantization_params(input_details)
    )

  def test_get_input_tensor_names(self):
    input_tensor_names = tfl_interpreter_utils.get_input_tensor_names(
        self._test_model_path
    )
    self.assertEqual(
        input_tensor_names,
        ["serving_default_conv2d_input:0"],
    )

  def test_get_output_tensor_names(self):
    output_tensor_names = tfl_interpreter_utils.get_output_tensor_names(
        self._test_model_path
    )
    self.assertEqual(
        output_tensor_names,
        ["StatefulPartitionedCall:0"],
    )

  def test_get_constant_tensor_names(self):
    const_tensor_names = tfl_interpreter_utils.get_constant_tensor_names(
        self._test_model_path
    )
    self.assertEqual(
        set(const_tensor_names),
        set([
            "sequential/conv2d/Conv2D",
            "sequential/conv2d/Relu;sequential/conv2d/BiasAdd;sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp",
            "arith.constant",
            "arith.constant1",
            "arith.constant2",
            "arith.constant3",
        ]),
    )


class TflUtilsQuantizedModelTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "conv_fc_mnist_srq_a8w8.tflite"
    )
    self._signature_input_data = {
        "conv2d_input": np.random.rand(1, 28, 28, 1).astype(np.float32)
    }

  def test_has_quantization_params(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    input_details = tfl_interpreter.get_input_details()[0]
    self.assertTrue(
        tfl_interpreter_utils.has_quantization_params(input_details)
    )

  def test_invoke_interpreter_signature(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    signature_output = tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data
    )
    self.assertEqual(tuple(signature_output["dense_1"].shape), (1, 10))

    # Assert the input data is not modified in-place b/353340272.
    self.assertEqual(
        self._signature_input_data["conv2d_input"].dtype, np.float32
    )

  def test_invoke_interpreter_signature_with_quantized_input(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    # Get the expected result with float input (input quantization happens
    # inside the model).
    expected_output = tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data
    )

    # Quantize the input data.
    input_details = tfl_interpreter.get_input_details()[0]
    scale, zero_point = input_details["quantization"]
    input_data = self._signature_input_data["conv2d_input"]
    quantized_input_data = (
        (input_data / scale + zero_point).round().astype(input_details["dtype"])
    )
    quantized_input = {"conv2d_input": quantized_input_data}

    # Get the result with quantized input.
    actual_output = tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, quantized_input
    )

    np.testing.assert_array_equal(
        expected_output["dense_1"], actual_output["dense_1"]
    )


class TflUtilsMultiSignatureModelTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "two_signatures.tflite"
    )
    self._signature_input_data = {"x": np.array([2.0]).astype(np.float32)}

  def test_create_tfl_interpreter(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    self.assertIsNotNone(tfl_interpreter)

  def test_get_input_tensor_names(self):
    signature_name = "add"
    input_tensor_names = tfl_interpreter_utils.get_input_tensor_names(
        self._test_model_path, signature_name
    )
    self.assertEqual(
        input_tensor_names,
        ["add_x:0"],
    )

    signature_name = "multiply"
    input_tensor_names = tfl_interpreter_utils.get_input_tensor_names(
        self._test_model_path, signature_name
    )
    self.assertEqual(
        input_tensor_names,
        ["multiply_x:0"],
    )

  def test_get_output_tensor_names(self):
    signature_name = "add"
    input_tensor_names = tfl_interpreter_utils.get_output_tensor_names(
        self._test_model_path, signature_name
    )
    self.assertEqual(
        input_tensor_names,
        ["PartitionedCall:0"],
    )

    signature_name = "multiply"
    input_tensor_names = tfl_interpreter_utils.get_output_tensor_names(
        self._test_model_path, signature_name
    )
    self.assertEqual(
        input_tensor_names,
        ["PartitionedCall_1:0"],
    )

  def test_get_constant_tensor_names(self):
    subgraph0_const_tensor_names = (
        tfl_interpreter_utils.get_constant_tensor_names(
            self._test_model_path, 0
        )
    )
    self.assertEqual(subgraph0_const_tensor_names, ["Add/y"])

    subgraph1_const_tensor_names = (
        tfl_interpreter_utils.get_constant_tensor_names(
            self._test_model_path, 1
        )
    )
    self.assertEqual(subgraph1_const_tensor_names, ["Mul/y"])

  def test_get_signature_main_subgraph_index(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    add_subgraph_index = (
        tfl_interpreter_utils.get_signature_main_subgraph_index(
            tfl_interpreter, "add"
        )
    )
    self.assertEqual(add_subgraph_index, 0)
    multiply_subgraph_index = (
        tfl_interpreter_utils.get_signature_main_subgraph_index(
            tfl_interpreter, "multiply"
        )
    )
    self.assertEqual(multiply_subgraph_index, 1)

  def test_get_tensor_data(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    # Invoke the ADD signature.
    tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data, "add"
    )
    output_details = {"index": 2, "quantization_parameters": {"scales": []}}
    output_data = tfl_interpreter_utils.get_tensor_data(
        tfl_interpreter, output_details, subgraph_index=0
    )  # The ADD signature is in the first subgraph.
    self.assertEqual(output_data, [12.0])  # 10 + 2

    # Invoke the MULTIPLY signature.
    tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data, "multiply"
    )
    output_data = tfl_interpreter_utils.get_tensor_data(
        tfl_interpreter, output_details, subgraph_index=1
    )  # The Multiply signature is in the second subgraph.
    self.assertEqual(output_data, [20.0])  # 10 * 2

  def test_get_tensor_name_to_content_map(self):
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    # Invoke all signatures.
    tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data, "multiply"
    )
    tfl_interpreter_utils.invoke_interpreter_signature(
        tfl_interpreter, self._signature_input_data, "add"
    )

    # Test tensors belonging to the ADD signature.
    add_subgraph_index = (
        tfl_interpreter_utils.get_signature_main_subgraph_index(
            tfl_interpreter, "add"
        )
    )
    add_tensor_content = tfl_interpreter_utils.get_tensor_name_to_content_map(
        tfl_interpreter, add_subgraph_index
    )

    add_input_content = add_tensor_content["add_x:0"]
    self.assertSequenceAlmostEqual(
        self._signature_input_data["x"].flatten(), add_input_content.flatten()
    )
    weight_content = add_tensor_content["Add/y"]
    self.assertEqual(weight_content, 10)
    add_output_content = add_tensor_content["PartitionedCall:0"]
    self.assertEqual(add_output_content, [12.0])

    # Test tensors belonging to the MULTIPLY signature.
    multiply_subgraph_index = (
        tfl_interpreter_utils.get_signature_main_subgraph_index(
            tfl_interpreter, "multiply"
        )
    )
    mul_tensor_content = tfl_interpreter_utils.get_tensor_name_to_content_map(
        tfl_interpreter, multiply_subgraph_index
    )
    multiply_input_content = mul_tensor_content["multiply_x:0"]
    self.assertSequenceAlmostEqual(
        self._signature_input_data["x"].flatten(),
        multiply_input_content.flatten(),
    )
    weight_content = mul_tensor_content["Mul/y"]
    self.assertEqual(weight_content, 10)
    multiply_output_content = mul_tensor_content["PartitionedCall_1:0"]
    self.assertEqual(multiply_output_content, [20.0])


class TflUtilsIntegerInputModelTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "toy_model_with_kv_cache_multi_signature.tflite"
    )

  def test_random_integer_input_data(self):
    test_data = tfl_interpreter_utils.create_random_normal_input_data(
        self._test_model_path
    )
    self.assertEqual(test_data["signature_1"][0]["cache_0"].dtype, np.float32)
    self.assertEqual(test_data["signature_1"][0]["cache_1"].dtype, np.float32)
    self.assertEqual(test_data["signature_1"][0]["positions"].dtype, np.int32)
    self.assertEqual(test_data["signature_1"][0]["tokens"].dtype, np.int32)

  def test_create_random_integers_defaults(self):
    rng = np.random.default_rng(0)

    # Test int8 defaults
    dtype = np.dtype(np.int8)
    data = tfl_interpreter_utils._create_random_integers(rng, (1000,), dtype)
    self.assertEqual(data.dtype, dtype)
    self.assertGreaterEqual(data.min(), np.iinfo(dtype).min)
    self.assertLessEqual(data.max(), np.iinfo(dtype).max)
    # Check if we cover a reasonable range (probability check)
    self.assertGreater(data.max(), 0)
    self.assertLess(data.min(), 0)

    # Test int16 defaults
    dtype = np.dtype(np.int16)
    data = tfl_interpreter_utils._create_random_integers(rng, (1000,), dtype)
    self.assertEqual(data.dtype, dtype)
    self.assertGreaterEqual(data.min(), np.iinfo(dtype).min)
    self.assertLessEqual(data.max(), np.iinfo(dtype).max)

    # Test int32 defaults
    dtype = np.dtype(np.int32)
    data = tfl_interpreter_utils._create_random_integers(rng, (1000,), dtype)
    self.assertEqual(data.dtype, dtype)
    self.assertGreaterEqual(data.min(), np.iinfo(dtype).min)
    self.assertLessEqual(data.max(), np.iinfo(dtype).max)

    # Test int64 defaults
    dtype = np.dtype(np.int64)
    data = tfl_interpreter_utils._create_random_integers(rng, (1000,), dtype)
    self.assertEqual(data.dtype, dtype)
    self.assertGreaterEqual(data.min(), np.iinfo(dtype).min)
    self.assertLessEqual(data.max(), np.iinfo(dtype).max)

    # Test explicit min_max_range
    dtype = np.dtype(np.int32)
    data = tfl_interpreter_utils._create_random_integers(
        rng, (1000,), dtype, min_value=-10, max_value=10
    )
    self.assertGreaterEqual(data.min(), -10)
    self.assertLess(data.max(), 10)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""test for quantize tensor."""

import pathlib
import numpy as np
from absl.testing import parameterized
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import quantize_tensor
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("..")


class QuantizeTensorTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._orig_test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/insert_dequant_test.tflite"
    )
    self._model = tfl_flatbuffer_utils.read_model(self._orig_test_model_path)

  def test_quantize_constant_tensor(self):
    """test quantizing a constant tensor."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    data = np.ones([1, 112, 112, 32], dtype=np.int8)
    ret = quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            7,
            model.operatorCodes,
            model.buffers,
            subgraph,
            -1,
            [4],
            qtyping.UniformQuantParams(
                8, None, np.ones(1), np.ones(1), True, data
            ),
        )
    )
    self.assertEqual(ret.op_id, 0)
    self.assertEqual(ret.num_ops_added, 0)
    self.assertListEqual(
        np.array(model.buffers[8].data).tolist(), data.flatten().tolist()
    )
    quant_param = subgraph.tensors[7].quantization
    self.assertListEqual(np.array(quant_param.scale).tolist(), [1])
    self.assertEqual(np.array(quant_param.zeroPoint).tolist(), [1])
    self.assertEqual(quant_param.quantizedDimension, 0)

  def test_quantize_activation_tensor(self):
    """test quantizing an activation tensor."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    ret = quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            4,
            model.operatorCodes,
            model.buffers,
            subgraph,
            1,
            [3],
            qtyping.UniformQuantParams(
                8, None, np.array([22]), np.array([127])
            ),
        )
    )
    self.assertEqual(ret.op_id, 0)
    self.assertEqual(ret.num_ops_added, 0)
    quant_param = subgraph.tensors[4].quantization
    self.assertListEqual(np.array(quant_param.scale).tolist(), [22])
    self.assertListEqual(np.array(quant_param.zeroPoint).tolist(), [127])
    self.assertEqual(quant_param.quantizedDimension, 0)

  def test_quantize_tensor_with_per_channel_quantization(self):
    """test quantizing an activation tensor."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    ret = quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            4,
            model.operatorCodes,
            model.buffers,
            subgraph,
            1,
            [3],
            qtyping.UniformQuantParams(8, 3, np.ones([22]), np.zeros([22])),
        )
    )
    self.assertEqual(ret.op_id, 0)
    self.assertEqual(ret.num_ops_added, 0)
    quant_param = subgraph.tensors[4].quantization
    self.assertListEqual(
        np.array(quant_param.scale).tolist(), np.ones([22]).tolist()
    )
    self.assertListEqual(
        np.array(quant_param.zeroPoint).tolist(), np.zeros([22]).tolist()
    )
    self.assertEqual(quant_param.quantizedDimension, 3)

  def test_quantize_tensor_with_nonlinear_quantization(self):
    """test quantizing an activation tensor with non-linear quantization."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            4,
            model.operatorCodes,
            model.buffers,
            subgraph,
            1,
            [3],
            qtyping.NonLinearQuantParams(16, None),
        )
    )
    self.assertEqual(
        subgraph.tensors[4].type, schema_py_generated.TensorType.FLOAT16
    )

  def test_blockwise_quantization_with_zero_point(self):
    """Test blockwise quantization with explicit zero point."""
    subgraph = self._model.subgraphs[0]
    model = self._model
    tensor_wh = 112
    test_tensor_id = 7
    data = np.ones([1, tensor_wh, tensor_wh, 32]).astype(np.int8)
    quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            tensor_id=test_tensor_id,
            op_codes=model.operatorCodes,
            buffers=model.buffers,
            subgraph=subgraph,
            producer=1,
            consumers=[3],
            quant_params=qtyping.UniformQuantParams(
                num_bits=8,
                quantized_dimension=None,
                scale=np.ones([1, tensor_wh, tensor_wh, 1]).astype(np.float16),
                zero_point=np.zeros([1, tensor_wh, tensor_wh, 1]),
                symmetric=True,
                quantized_data=data,
                block_size=32,
            ),
        )
    )
    quant_param = subgraph.tensors[test_tensor_id].quantization
    self.assertEqual(
        quant_param.detailsType,
        schema_py_generated.QuantizationDetails.BlockwiseQuantization,
    )
    self.assertEqual(quant_param.details.blockSize, 32)
    # Check if the scale and zero point tensors are inserted correctly.
    self.assertEqual(quant_param.details.scales, 9)
    # So far we don't have zero point in blockwise quantization.
    self.assertEqual(quant_param.details.zeroPoints, -1)

  def test_int4_constant_packed_correctly(self):
    subgraph = self._model.subgraphs[0]
    model = self._model
    data = np.array(
        [
            0x0,
            0x1,
            0x2,
            0x3,
            0x4,
            0x5,
            0x6,
            0x7,
            0x8,
            0x9,
            0xA,
            0xB,
            0xC,
            0xD,
            0xE,
        ],
        dtype=np.int8,
    )
    expected = np.array([0x10, 0x32, 0x54, 0x76, 0x98, 0xBA, 0xDC, 0x0E])
    ret = quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            tensor_id=7,
            op_codes=model.operatorCodes,
            buffers=model.buffers,
            subgraph=subgraph,
            producer=-1,
            consumers=[4],
            quant_params=qtyping.UniformQuantParams(
                4, None, np.ones(1), np.ones(1), True, data
            ),
        )
    )
    self.assertEqual(ret.op_id, 0)
    self.assertEqual(ret.num_ops_added, 0)
    np.testing.assert_array_equal(model.buffers[8].data, expected)
    quant_param = subgraph.tensors[7].quantization
    np.testing.assert_array_equal(quant_param.scale, [1])
    np.testing.assert_array_equal(quant_param.zeroPoint, [1])
    self.assertEqual(quant_param.quantizedDimension, 0)

  @parameterized.named_parameters(
      dict(
          testcase_name="int5",
          num_bits=5,
      ),
      dict(
          testcase_name="int2",
          num_bits=2,
      ),
  )
  def test_int_constant_not_packed(self, num_bits):
    subgraph = self._model.subgraphs[0]
    model = self._model
    tensor_id = 7
    data = np.array([0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7], dtype=np.int8)
    expected = np.array([0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7])
    ret = quantize_tensor.quantize_tensor(
        transformation_utils.TransformationInput(
            tensor_id=tensor_id,
            op_codes=model.operatorCodes,
            buffers=model.buffers,
            subgraph=subgraph,
            producer=-1,
            consumers=[4],
            quant_params=qtyping.UniformQuantParams(
                num_bits=num_bits,
                quantized_dimension=None,
                scale=np.ones(1),
                zero_point=np.ones(1),
                symmetric=True,
                quantized_data=data,
            ),
        )
    )
    self.assertEqual(ret.op_id, 0)
    self.assertEqual(ret.num_ops_added, 0)
    np.testing.assert_array_equal(model.buffers[8].data, expected)
    quant_param = subgraph.tensors[tensor_id].quantization
    np.testing.assert_array_equal(quant_param.scale, [1])
    np.testing.assert_array_equal(quant_param.zeroPoint, [1])
    self.assertEqual(quant_param.quantizedDimension, 0)


if __name__ == "__main__":
  absltest.main()
